/*  GDFutils.C
    utility routines
*/

#ifndef UTILS

#define UTILS = 1;    /* dummy for ifndef */

/* prototypes */


void CheckRAM(void *ptr);
     /* verify that a pointer points to RAM */
int  strpos(char *string, char what);
     /* locate the index position of character <what> in <string> */
char *strupper(char *string);
     /* make <string> all upper case */
char *strchrInsert(char inchar, char *string, int I);
     /* insert character <inchar> into <string> at position <I> */
void strRemoveLspace(char *string);
     /* remove leading blanks from <string> */
void strRemoveRspace(char *string);
     /* remove trailing blanks from <string> */
void strRemoveDspace(char *string);
     /* replace multiple blanks with single blanks in <string> */
void strRemoveAllspace(char *string);
     /* remove all blanks from <string> */
int  maxInteger(int i1,int i2);
     /* return maximum of two integers */
int  minInteger(int i1,int i2);
     /* return minimum of two integers */
char *ValidateDOSfileName(char *initname, char trim);
     /* validate and correct a potential DOS filename */
int  Verify(int verify, char *string);
     /* make sure <string> conforms to a verification list */
int  FORTRANwidth(char *format);
     /* find the logical width of a FORTRAN <format> */


/* implimentation */

char DOSfilename[100];    /*  */

void CheckRAM(void *ptr)
     /* verify that a pointer points to RAM */
{
    if (ptr == NULL) {
      alert('A',"Sorry - run out of RAM","ASEG-GDF will abort");
      exit(3);
    }
}  /* end CheckRAM */

int strpos(char *string, char what)
    /* locate the index position of character <what> in <string> */
{ int P,L;
  if (what=='\0')  /* null character is an error */
    return -1;
  else {
    P = 0;
    L = strlen(string);
    while ((P < L) && (string[P] != what)) P++;
    if (string[P]==what)
      return P;
    else
      return -1;   /* character not found */
  }
}  /* end strpos */

char *strupper(char *string)
     /* make <string> all upper case */
{    int P;
    for (P = 0; P < strlen(string); P++)
      if ((string[P] >= 'a') && (string[P] <= 'z')) string[P] -= 32;
    return string;
}  /* end strupper */

char *strchrInsert(char inchar, char *string, int I)
     /* insert character <inchar> into <string> at position <I> */
{   int P;
    for (P = strlen(string); P > I; P--) string[P] = string[P-1];
    string[I] = inchar;
    return string;
}  /* end strchrInsert */

void strRemoveLspace(char *string)
     /* remove leading blanks from <string> */
{   int len,I,P;
    len = strlen(string);
    I = 0;
    while (string[I] == ' ') I++;
    if (I > 0) for (P = I; P <= len; P++) string[P-I] = string[P];
}  /* end strRemoveLspace */

void strRemoveRspace(char *string)
     /* remove trailing blanks from <string> */
{   int I;
    I = strlen(string) - 1;
    while (string[I] == ' ') I--;
    string[I+1] = '\0';
}  /* end strRemoveRspace */

void strRemoveDspace(char *string)
     /* replace multiple blanks with single blanks in <string> */
{   int I,J;
    int len;
    len = strlen(string) - 1;
    for (I = 0, J = 1; I < len; I++) {
      if ((string[I] == ' ') && (string[I+1] == ' ')) J--;
      string[J++] = string[I+1];
    }
    string[J] = '\0';
}  /* end strRemoveDspace */

void strRemoveAllspace(char *string)
     /* remove all blanks from <string> */
{   int I,P;
    I = 0;
    while ((string[I] != ' ') && (string[I] != '\0')) I++;
    if (string[I] == ' ') {
      P = I;
      while (string[I] != '\0') {
        if (string[I] == ' ') I++;
        string[P++] = string[I++];
      }
      string[P] = '\0';
    }
}  /* end strRemoveAllspace */

int  maxInteger(int i1,int i2)
     /* return maximum of two integers */
{   if (i1 > i2)
      return i1;
    else
      return i2;
}  /* end maxInteger */
int  minInteger(int i1,int i2)
     /* return minimum of two integers */
{   if (i1 > i2)
      return i2;
    else
      return i1;
}  /* end minInteger */

char *ValidateDOSfileName(char *initname, char trim)
     /* validate and correct a potential DOS filename */
{   int I,L;
    strcpy(DOSfilename,initname);  /* do not alter the original */
    L = strlen(DOSfilename);
       /* replace blanks with underscores */
    while ((I = strpos(DOSfilename,' ')) >= 0) DOSfilename[I] = '_';
       /* replace invalid characters with '^' */
    for (I = 0; I < L; I++) {
      if (strpos("@ABCDEFGHIJKLMNOPQRSTUVWXYZ^_{}0123456789.!#$%&()~",toupper(DOSfilename[I])) == -1)
        DOSfilename[I] = '^';
    }
       /* trim name to <trim> characters, removing extension if requested */
    if ((trim == '.') && ((I = strpos(DOSfilename,'.')) >= 0))
      DOSfilename[I] = '\0';
    else if (trim != '\0') DOSfilename[I] = '\0';
    if (DOSfilename[0] == '\0') strcpy(DOSfilename,"BAD_NAME");

    return DOSfilename;
}  /* end ValidateDOSfileName */

int  Verify(int verify, char *string)
     /* make sure <string> conforms to a verification list
        verify is a set of 16 bit flags
        1 = remove all spaces
        2 = make upper case
        4 = DOS filename
        8 = Unix filename
       16 = FORTRAN input format
     */
{   int I,len,P;         /* indexes */
    int err = 0;         /* return error number */
    int dot;             /* counts decimal dots */
    int Ftype;           /* index to format type */

    if (verify & 1) strRemoveAllspace(string);
    if (verify & 2) string = strupper(string);   /* make UPPER case */
    if (verify & 4) {                            /* test DOS filename */
      strcpy(string,ValidateDOSfileName(string,'.'));
      err = 3*(string == "BAD_NAME");
      }
/*    if (verify & 8) ;                             test UNIX filename */
    if (verify & 16) {                           /* FORTRAN format check */
      if ((P = strpos("IALXDEFG",string[0])) > -1) {  /* must start with one */
        Ftype = P;
        len = strlen(string);
        I = 1; dot = 0;
        while ((I < len) && (P > -1)) {
          P = strpos(".0123456789",string[I++]);
          if (P == 0) {                  /* a dot */
            if (I == 2) err = 7;         /* must have a format width */
            else if (Ftype < 4) err = 7; /* no dot in IALX */
            else dot++;                  /* only one dot allowed in DEFG */
          }
        }
      }
      if ((len < 2) || (P == -1) || (dot > 1)) err = 7;
    }
    if (verify & 32) {
      strRemoveLspace(string);
      strRemoveRspace(string);
      strRemoveDspace(string);
    }
    return err;
}  /* end Verify */


int FORTRANwidth(char *format)
     /* find the logical width of a FORTRAN <format> */
{   int I = 1;
    int value = 0;
    int len;
    len = strlen(format);
       /* test data type */
    if (strpos("IFLDAGEX",format[0]) == -1) return 0;  /* bad format */
       /* find data width */
    while ((I < len) && (strpos("0123456789",format[I]) > -1))
      value = value*10 + format[I++] - '0';

    return value;
}  /* end FORTRANwidth */

#endif