/* ScreenIO.C
      screen handling constants, parameters and functions
      keyboard constants
*/

#ifndef SVERTICAL            /* prevents duplicate includes */

   /* box-drawing characters */
#define SVERTICAL   179      /* single lines */
#define SHORIZONTAL 196
#define SLOWLEFT    192
#define SLOWRIGHT   217
#define STOPLEFT    218
#define STOPRIGHT   191
#define STOP        194
#define SBOTTOM     193
#define SLEFT       195
#define SRIGHT      180
#define SCROSS      197

#define DVERTICAL   186      /* double lines */
#define DHORIZONTAL 205
#define DLOWLEFT    200
#define DLOWRIGHT   188
#define DTOPLEFT    201
#define DTOPRIGHT   187
#define DTOP        203
#define DBOTTOM     202
#define DLEFT       204
#define DRIGHT      185
#define DCROSS      206

#define DSTOP       209      /* double+single line joins*/
#define DSBOTTOM    207
#define DSLEFT      199
#define DSRIGHT     182
#define SDTOP       210
#define SDBOTTOM    208
#define SDLEFT      198
#define SDRIGHT     181
#define DSCROSSHORZ 215
#define DSCROSSVERT 216

   /* keyboard key-codes */
#define BLOTCH      219
#define BACKSPACE     8
#define ENTER        13
#define ESC          27
#define BELL          7
#define CUT          20
#define COPY          3
#define PASTE        16
#define HOME         71
#define END          79
#define PAGEUP       73
#define PAGEDN       81
#define LARROW       75
#define RARROW       77
#define DNARROW      80
#define UPARROW      72
#define TAB           9
#define SHIFTTAB     15
#define INS          82
#define DEL          83
#define F8           66
#define CTRLLARROW  115
#define CTRLRARROW  116
#define EXTKEY        0
#define ALT           1      /* psudo code */

char buff[200];        /* tempory screen text buffer */

   /* defaults replaced when GetScreen called */
int screenRight  = 80;
int screenBottom = 25;

struct boxes {         /* parameters associated with displayed boxes */
     char style;
     int  left,right,top,bottom;
     int  Tcol,Bcol;
} box[9];

struct KeepArea {      /* holds info to replace screen image */
     int width,height;      /* width and height */
     int l,t;               /* position on screen of left top*/
     int WindowArea[1000];  /* holds covered screen image */
} AreaA;
int alertOn = 0;       /* true if an alert box has been left visible */

int setx[9],sety[9];   /* the positions of grid lines in boxes (not saved)
                          set up in eg GDFLISTS.ListOn */
static unsigned char charh,charv,charx, charl,charr,chart,charb;
                       /* char set obtained from box type, used for grids */
char alertAns[80];     /* returned string response */

int  getx(void);
int  gety(void);
/* getx and gety get the cursor column and row window_relative location */

void SetTextColour(int colfor, int colback);
     /* sets the forground and background text screen colours */
void GetScreen(void);
     /* stores screen parameters in screenInfo */
void SetWindow(int fromL, int fromT, int fromR, int fromB);
     /* sets the display window
        <from[LTRB]> are the margins from left,top,right,bottom */
void ClearWindow(void);
     /* clears the current window */
void ClearLine(int colb, int coltf, int coltb);
     /* clears the current line in window in colour <colb>
        and sets colours */
void ClearScreen(void);
     /* sets window back to all and clears it white on black */
int  WhereY(void);
     /* returns absolute screen row */
void ClearTick(int tick, int colf, int colb);
     /* sets window to all but top & bottom lines and clears it <colf> on <colb>
        with optional tick marks above and below */

void PrintChar(int x, int y, char *format, char *text);
     /* displays <text> using <format> in column <x> of line <y> */
void PrintInt(int x, int y, char *format, long value);
     /* displays <value> using <format> in column <x> of line <y> */
void PrintSetup(unsigned char posn, int y, int colf, int colb, int textlen, struct text_info winInfo);
     /* prepares display for printing <textlen> chars in <winInfo>
     in the <posn> ['L'eft,'C'entre,'R'ight] of line <y>
     with colour <colf> on background colour <colb>
     if <posn> = lowercase [or > 128] then clear line prior to printing */
void PrintCharWindow(unsigned char posn, int y, int colf, int colb, char *text);
     /* displays <text> in current window
     in the <posn> ['L'eft,'C'entre,'R'ight] of line <y>
     with colour <colf> on background colour <colb>
     if <posn> = lowercase then clear line prior to printing
     afterwards it restores the window */
void PrintCharScreen(unsigned char posn, int y, int colf, int colb, char *text);
     /* same as PrintCharWindow but on original screen */
void PrintIntScreen(unsigned char posn, int y, int colf, int colb, long value);
     /* displays <value> on screen at <posn> on line <y>
        as for PrintCharScreen */
void Print3IntScreen(unsigned char posn,int y, int colf, int colb, int V1,int V2,int V3, int W1,int W2,int W3);
     /* displays V1, V2 and V3 on screen line <y> starting on column <posn>
        with colour <colf> on background colour <colb>
        with widths <W1,W2,W3> */

char *alert(char alertType, char *note1, char *note2);
     /* displays a box with message(s) and may wait for a response */
int  alertYesNo(char *note1, char *note2);
     /* displays a box with message(s) and waits for a Yes/No response */
void alertClear(void);
     /* replaces a box area with the hidden screen image */
void DrawWinBox(int boxn, int L, int T, int R, int B, char style, char clr, int Tcolour, int Bcolour);
     /* displays a box and makes the inside the text window */
void DrawBox(int L,int T, int R,int B,
             char vert,char horz, char topl,char topr, char lowl,char lowr);
     /* draws a box with corners at screen <L>eft/<T>op and <R>ight/<B>ottom
        using the characters listed */
void DrawSBox(int L, int T, int R, int B);
     /* displays a box using single lines */
void DrawDBox(int L, int T, int R, int B);
     /* displays a box using double lines */
void SetLines(int boxn, char style, int internal);
     /* sets the characters to use in grid drawing
        <boxn>  is the box requiring a grid - with Single or Double lines
        <style> is the grid style - Single or Double lined
        <internal> is true if internal top edge */
void DrawGrid(int boxn, int startx,int starty, int delx,int dely, int style);
     /* draws a grid of lines inside box <boxn> */
void DrawGridSet(int boxn, char style);
     /* draws a grid of lines inside box <boxn> as specified in setx & sety */

struct text_info screenInfo;


/* implimentation */

void textcolor(int newcolor);
void textbackground(int colback);
void window(int left, int top, int right, int bottom);
void gettextinfo(struct text_info *r);
int cputs(const char *str);
void textattr(int newattr);
int puttext(int left, int top, int right, int bottom, void *source);
int gettext(int left, int top, int right, int bottom, void *destin);


/* getx and gety get the cursor column and row window_relative location */
#if (PLATFORM == DOS)
int  getx(void)
{   return wherex();
}
int  gety(void)
{   return wherey();
}
#elif (PLATFORM == UNIX)
int  getx(void)
{   int c,r;  /* cursor column,row location */
    getyx(stdscr,r,c);
    return c;
}
int  gety(void)
{   int c,r;  /* cursor column,row location */
    getyx(stdscr,r,c);
    return r;
}
#endif

void SetTextColour(int colfor, int colback)
     /* sets the forground and background text screen colours */
{
    textcolor(colfor);
    textbackground(colback);
}

void GetScreen(void)
     /* stores screen parameters in screenInfo */
{
  gettextinfo(&screenInfo);
  screenBottom = screenInfo.screenheight;
  screenRight  = screenInfo.screenwidth;
}


void SetWindow(int fromL, int fromT, int fromR, int fromB)
     /* sets the display window
        <from[LTRB]> are the margins from left,top,right,bottom */
{
    window(1+fromL,1+fromT,screenInfo.screenwidth-fromR,screenInfo.screenheight-fromB);
}

void ClearWindow(void)
     /* clears the current window */
{
    clrscr();
}

void ClearLine(int colb, int coltf, int coltb)
     /* clears the current line in window in colour <colb>
        and sets colours */
{
    gotoxy(1,gety());
    SetTextColour(coltf,colb);
    clreol();
    SetTextColour(coltf,coltb);
}

void ClearScreen(void)
     /* sets window back to all and clears it white on black */
{
    SetWindow(0,0,0,0);
    SetTextColour(LIGHTGRAY,BLACK);
    clrscr();
}

int WhereY(void)
     /* returns absolute screen row */
{   struct text_info winInfo;
    gettextinfo(&winInfo);
    return gety() + winInfo.wintop - 1;
}

void ClearTick(int tick, int colf, int colb)
     /* sets window to all but top & bottom lines and clears it <colf> on <colb>
        with optional tick marks above and below */
{   int I,Y;
    int width,height;    /* screen width and height */
    width = screenInfo.screenwidth;
    height= screenInfo.screenheight;

    SetTextColour(colf,colb);
       /* wipe window clean, excluding Menu line */
    SetWindow(0,1,0,0);
    clrscr();
       /* if tick marks required */
    if (tick > 5) {
      for (Y = 2; Y < height; Y += height-3) { /* top and bottom sets */
        window(1,Y,width,Y+1);
        for (I = 5; I <= width; I++)
          if ((I % tick))
            putch(' ');
          else {
            printf("%3i |",I);
            I += 4;
          }
      }
    }
       /* set data display area of screen */
    SetWindow(0,2,0,2);
}  /* end ClearTick */


void PrintChar(int x, int y, char *format, char *text)
     /* displays <text> using <format> in column <x> of line <y> */
{
    gotoxy(x,y);
       /* print message */
    sprintf(buff,format,text);
    cputs(buff);
}  /* end PrintChar */

void PrintInt(int x, int y, char *format, long value)
     /* displays <value> using <format> in column <x> of line <y> */
{
    gotoxy(x,y);
       /* print message */
    sprintf(buff,format,value);
    cputs(buff);
}  /* end PrintInt */


void PrintSetup(unsigned char posn, int y, int colf, int colb, int textlen, struct text_info winInfo)
     /* prepares display for printing <textlen> chars in <winInfo>
     in the <posn> ['L'eft,'C'entre,'R'ight] of line <y>
     with colour <colf> on background colour <colb>
     if <posn> = lowercase [or > 128] then clear line prior to printing */
{ int x;                             /* screen column */
     /* set requested colours */
  SetTextColour(colf,colb);
     /* clear line if requested */
  if (y > winInfo.winbottom) y = winInfo.winbottom;
  gotoxy(1,y);
  if ((posn > 'Z') || (posn > 128) || (textlen == 0)) clreol();
     /* print message */
  switch (toupper(posn)) {
    case 'L' : x = 1; break;
    case 'C' : x = (winInfo.winright - winInfo.winleft + 3 - textlen)/2;
               break;
    case 'R' : x = winInfo.winright - textlen;
               break;
    default  : if (posn > 128) x = 256-posn; else x = posn;
               if (x <= (winInfo.winright - textlen)) break; /* test line overflow */
               x = winInfo.winright - textlen;
  }
  if (x < 1) x = 1;   /* text is wider than screen ! */
  gotoxy(x,y);
}  /* end PrintSetup */

void PrintCharWindow(unsigned char posn, int y, int colf, int colb, char *text)
     /* displays <text> in current window
        in the <posn> ['L'eft,'C'entre,'R'ight] of line <y>
        with colour <colf> on background colour <colb>
        if <posn> = lowercase then clear line prior to printing */
{
  struct text_info winInfo;
  gettextinfo(&winInfo);
     /* display text */
  PrintSetup(posn, y, colf,colb, strlen(text), winInfo);
  cputs(text);
     /* restore original colour */
  textattr(winInfo.attribute);
}  /* end PrintCharWindow */


void PrintCharScreen(unsigned char posn, int y, int colf, int colb, char *text)
     /* displays <text> on screen then restores calling window
     in the <posn> ['L'eft,'C'entre,'R'ight] of line <y>
     with colour <colf> on background colour <colb>
     if <posn> = lowercase then clear line prior to printing */
{
  struct text_info winInfo;          /* to hold caller's window info */
  gettextinfo(&winInfo);
     /* make screen = original screen */
  window(screenInfo.winleft,screenInfo.wintop,screenInfo.winright,screenInfo.winbottom);
  textattr(screenInfo.attribute);
     /* display text */
  PrintSetup(posn, y, colf,colb, strlen(text), screenInfo);
  cputs(text);
     /* restore caller's window */
  window(winInfo.winleft,winInfo.wintop,winInfo.winright,winInfo.winbottom);
  textattr(winInfo.attribute);
  gotoxy(winInfo.curx,winInfo.cury);
}  /* end PrintCharScreen */


void PrintIntScreen(unsigned char posn, int y, int colf, int colb, long value)
     /* displays <value> on screen at <posn> on line <y>
        then as for PrintCharScreen */
{   char val[40];
    sprintf(val,"%li",value);
    PrintCharScreen(posn, y, colf, colb, val);
}  /* end PrintIntScreen */

void Print3IntScreen(unsigned char posn,int y, int colf, int colb, int V1,int V2,int V3, int W1,int W2,int W3)
     /* displays V1, V2 and V3 on screen line <y> starting on column <posn>
        with colour <colf> on background colour <colb>
        with widths <W1,W2,W3> */
{   char val[120];
    sprintf(val,"%*d %*d %*d",W1,V1,W2,V2,W3,V3);
    PrintCharScreen(posn, y, colf, colb, val);
}  /* end Print3IntScreen */


char *alert(char alertType, char *note1, char *note2)
     /* displays a double_lined box with message(s) and waits for a response
        <alertType> = 'A' for any key
                      'C' for char
                      'S' for string
                      'L' for leave message on screen
        <note1> & <note2> are messages to display before accepting response */
{   int L;                 /* total number of lines of text and spaces */
    int L1,L2;             /* lengths of note1 & note2 to maximum width */
    int width,height;      /* width and height */
    int w,h;               /* position on screen of left top*/
    char alertChar;        /* char reponse */
    struct text_info winInfo;  /* caller's window */
    gettextinfo(&winInfo);

       /* trim messages to 5 lines each */
    L = 5*(screenRight-4);
    L1 = strlen(note1);
    if (L1 > L) { note1[L] = '\0'; L1 = L; }
    L2 = strlen(note2);
    if (alertType == 'S') L2 = L2 + 15;       /* allow for response */
    if (L2 > L) { note2[L] = '\0'; L2 = L; }
    L = L1/(screenRight-4) + L2/(screenRight-4);
       /* trim L1 & L2 to box window */
    if (L1+4 > screenRight) L1 = screenRight-4;
    if (L2+4 > screenRight) L2 = screenRight-4;
       /* determine box size  and centre it on screen */
    if (L1>L2) width = L1+4; else width = L2+4;
    height = L + 1 + 2*(L1>0) + 2*(L2>0);
    w = screenRight - width - 1;
    if (alertType != 'L') w = w/2;
    if (w<1) w = 1;
    h = (screenBottom - height)/2; if (h<1) h = 1;
       /* store covered screen image */
    gettext(w,h, w+width-1,h+height-1,AreaA.WindowArea);
       /* draw box and create window */
    DrawWinBox(7,w,h, w+width-1,h+height-1, 'S','c',BLACK,WHITE);
       /* write message(s) */
    if (L1) { gotoxy((width-L1)/2,1); cputs(note1); }
    if (L2) { gotoxy((width-L2)/2,gety()+2); cputs(note2); }
       /* wait for user to respond */
    switch (alertType) {
      case 'A' : ;
      case 'C' : alertChar = toupper(getch());
                 if (alertChar == 0) alertChar = getch(); /* discard double_key code */
                 alertAns[0] = alertChar;
                 alertAns[1] = '\0';
                 break;
      case 'S' : gets(alertAns);
    }
       /* restore screen image and caller's window */
    if (alertType == 'L') {
      alertAns[0] = '\0';
      AreaA.width  = width-1;
      AreaA.height = height-1;
      AreaA.l = w;
      AreaA.t = h;
      alertOn = 1;
      }
    else {
      puttext(w,h, w+width-1,h+height-1,AreaA.WindowArea);
      alertOn = 0;
      window(winInfo.winleft,winInfo.wintop,winInfo.winright,winInfo.winbottom);
      textattr(winInfo.attribute);
    }
    return alertAns;
}  /* end alert */


int  alertYesNo(char *note1, char *note2)
     /* displays a box with message(s) and waits for a Yes/No response */
{   char ans[4];
    strcpy(ans,alert('C', note1, note2));
    return (ans[0] == 'Y');
}  /* end alertYesNo */

void alertClear(void)
     /* replaces a box area with the hidden screen image */
{   if (alertOn) {
      puttext(AreaA.l,AreaA.t, AreaA.l+AreaA.width,AreaA.t+AreaA.height,AreaA.WindowArea);
      alertOn = 0;
    }
}  /* end alertClear */

void DrawWinBox(int boxn, int L,int T,int R,int B, char style, char clr, int Tcolour, int Bcolour)
     /* displays box <boxn> and makes the inside the text window
        box bounds are <L,R,T,B>
        <style> is single or double {'S','D'}
        <clr> = 'c' if window area is to be wiped
        <Tcolour> & <Bcolour> are text and background colours */
{   int screenR;
    if (R > screenRight) screenR = screenRight; else screenR = R;
    box[boxn].style = style;
    box[boxn].left  = L;
    box[boxn].right = screenR;
    box[boxn].top   = T;
    box[boxn].bottom= B;
    box[boxn].Tcol  = Tcolour;
    box[boxn].Bcol  = Bcolour;
    SetTextColour(Tcolour,Bcolour);
    if (clr=='c') {
      window(L,T, screenR,B);
      clrscr();
    }
    if (B<screenInfo.winbottom) window(L,T, screenR,B+1);
    else window(L,T, screenR+1,B);

    if (style == 'D') DrawDBox(1,1, R-L+1,B-T+1);
    else DrawSBox(1,1, R-L+1,B-T+1);

    window(L+1,T+1, screenR-1,B-1);
}  /* end DrawWinBox */

void DrawBox(int L,int T, int R,int B,
             char vert,char horz, char topl,char topr, char lowl,char lowr)
     /* draws a box
        with corners at screen <L>eft/<T>op and <R>ight/<B>ottom
        using text characters of
          <vert> for verticals
          <horz> for horizontals
          <topl> for top left corner
          <topr> for top right corner
          <lowl> for bottom left corner
          <lowr> for bottom right corner */
{   int I;
    int screenR;

    if (R > screenRight) screenR = screenRight; else screenR = R;
       /* draw verticals */
    for (I = T+1; I< B; I++) {
      gotoxy(L,I); putch(vert);
      if (R <= screenRight) { gotoxy(R,I); putch(vert); }
    }
       /* draw corners & horizontals */
    gotoxy(L,T); putch(topl);
    for (I = L+1; I< screenR; I++) putch(horz);
    if (R <= screenRight) putch(topr);
    else putch(horz);

    gotoxy(L,B); putch(lowl);
    for (I = L+1; I< screenR; I++) putch(horz);
    if (R <= screenRight) putch(lowr);
    else putch(horz);
}  /* end DrawBox */

void DrawSBox(int L, int T, int R, int B)
     /* displays a box using single lines
        with corners at screen <L>eft/<T>op and <R>ight/<B>ottom */
{
  DrawBox(L,T,R,B, SVERTICAL,SHORIZONTAL,STOPLEFT,STOPRIGHT,SLOWLEFT,SLOWRIGHT);
}  /* end DrawSBox */

void DrawDBox(int L, int T, int R, int B)
     /* displays a box using double lines
        with corners at screen <L>eft/<T>op and <R>ight/<B>ottom */
{
  DrawBox(L,T,R,B, DVERTICAL,DHORIZONTAL,DTOPLEFT,DTOPRIGHT,DLOWLEFT,DLOWRIGHT);
}  /* end DrawDBox */

void SetLines(int boxn, char style, int internal)
     /* sets the characters to use in grid drawing
        <boxn>  is the box requiring a grid - with Single or Double lines
        <style> is the grid style - Single or Double lined
        <internal> is true if internal top edge */
{
    if (style == 'D') {              /* grid needs double lines */
      charh = DHORIZONTAL;
      charv = DVERTICAL;
      charx = DCROSS;
      if (box[boxn].style == 'D') {  /* box has double lines */
        charl = DLEFT;
        charr = DRIGHT;
        charb = DBOTTOM;
        chart = DTOP;
        }
      else {                         /* box has single lines */
        charl = SDLEFT;
        charr = SDRIGHT;
        charb = SDBOTTOM;
        if (internal) chart = DTOP; else chart = SDTOP;
        }
      }
    else /* style == 'S' */ {        /* grid needs single lines */
      charh = SHORIZONTAL;
      charv = SVERTICAL;
      charx = SCROSS;
      if (box[boxn].style == 'D') {  /* box has double lines */
        charl = DSLEFT;
        charr = DSRIGHT;
        charb = DSBOTTOM;
        if (internal) chart = STOP; else chart = DSTOP;
        }
      else {                         /* box has single lines */
        charl = SLEFT;
        charr = SRIGHT;
        charb = SBOTTOM;
        chart = STOP;
      }
    }
}  /* end SetLines */

void DrawGrid(int boxn, int startx,int starty, int delx,int dely, int style)
     /* draws a grid of lines
        of <style> single or double
        inside box <boxn>.
        The grid starts at column <startx>, row <starty>
        and steps by <delx> columns, <dely> rows */
{   int w,h;
       /* get box's screen position */
    int L,R,T,B;
    L = box[boxn].left;
    R = box[boxn].right;
    T = box[boxn].top;
    B = box[boxn].bottom;
      /* make the box the current window and set the colours */
    window(L,T, R,B);
    SetTextColour(box[boxn].Tcol,box[boxn].Bcol);
       /* set the grid text characters to use */
    SetLines(boxn, style, 0);
       /* draw the grid horizontals and terminators */
    for (h = starty+1; h <= B-T; h = h+dely) {
      gotoxy(R-L+1,h); putch(charr);
      gotoxy(1,h); putch(charl);
      for (w = 2; w <= R-L; w++) putch(charh);
    }
       /* draw the grid verticals and terminators */
    for (w = startx+1; w <= R-L; w = w+delx) {
      gotoxy(w,B-T+1); putch(charb);
      for (h = 2; h <= B-T; h++) {
        gotoxy(w,h); putch(charv);
      }
    }
       /* draw the grid intersections */
    for (h = starty+1; h <= B-T; h = h+dely) {
      for (w = startx+1; w <= R-L; w = w+delx) {
        gotoxy(w,h); putch(charx);
      }
    }
       /* contract the window to inside the bounding box */
    window(L+1,T+1, R-1,B-1);
}  /* end DrawGrid */

void DrawGridSet(int boxn, char style)
     /* draws a grid of lines inside box <boxn> as specified in setx & sety
        with <style> single or double.
        Do not draw verticals across top strip */
{   int w,h, x,y;   /* indexes for setx and sety; column and row */
    int starty = 1; /* start y for verticals; not top strip if > 1 strip */
       /* get box's screen position */
    int L,R,T,B;
    L = box[boxn].left;
    R = box[boxn].right;
    T = box[boxn].top;
    B = box[boxn].bottom;
      /* make the box the current window and set the colours */
    window(L,T, R,B);
    SetTextColour(box[boxn].Tcol,box[boxn].Bcol);
       /* set the grid text characters to use */
    SetLines(boxn, style, 0);

       /* draw the grid horizontals and terminators */
    if (sety[0] > 1) {
      for (y = 0; (sety[y] > 1) && (sety[y] < B-T); y++) {
        h = sety[y] + 1;
        gotoxy(R-L+1,h); putch(charr);
        gotoxy(1,h); putch(charl);
        for (w = 2; w <= R-L; w++) putch(charh);
      }
      if (y > 1) {
        starty = sety[0]+1;
        SetLines(boxn, style, 1);  /* reset line joins */
      }
    }
       /* draw the grid verticals and terminators */
    if (setx[0] > 1) {
      for (x = 0; (setx[x] > 1) && (setx[x] < R-L); x++) {
        w = setx[x] + 1;
        gotoxy(w,B-T+1); putch(charb);
        gotoxy(w,starty); putch(chart);
        for (h = starty+1; h <= B-T; h++) {
          gotoxy(w,h); putch(charv);
        }
      }
       /* draw the grid intersections - if no verticals, then no intersections */
      for (y = (starty > 1); (sety[y] > 1) && (sety[y] < B-T); y++) {
        h = sety[y] + 1;
        for (x = 0; (setx[x] > 1) && (setx[x] < R-L); x++) {
          w = setx[x] + 1;
          gotoxy(w,h); putch(charx);
        }
      }
    }
       /* contract the window to inside the bounding box */
    window(L+1,T+1, R-1,B-1);
}  /* end DrawGridSet */

#endif